/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 *
 * $Id: analyze_header.c,v 1.26 2001/03/01 15:50:09 glauschwuffel Exp $
 *
 * $Log: analyze_header.c,v $
 * Revision 1.26  2001/03/01 15:50:09  glauschwuffel
 * HTTP header lines can now be ended by 0d,0a (specification) or just 0a
 * (altavista).
 *
 * Revision 1.25  2000/06/19 21:41:22  glauschwuffel
 * includes log.h
 *
 * Revision 1.24  2000/05/07 14:54:04  glauschwuffel
 * added include <ctype.h>
 *
 * Revision 1.23  2000/03/07 15:06:46  goldbach
 * added GPL info
 *
 * Revision 1.22  2000/02/21 15:15:03  gellert
 * removed the calls to strncpy...
 *
 * Revision 1.21  2000/01/25 15:03:25  gellert
 * Now the name and content of the actual headerline are processed
 * in the tct->inbuf. No copying to local variables (that are too
 * small for long lines (eg "Referer: thousands_of_bytes").
 *
 * Revision 1.20  2000/01/18 15:07:44  gellert
 * Now the http header lines are not copied, the input buffer
 * ist processed directly. This eleminates the problem of
 * short strings (or too long values of http entries).
 *
 * Revision 1.19  1999/12/01 17:53:02  gellert
 * bug-fix: field_name was a static pointer (used by all the threads!). ARGH!
 *
 * Revision 1.18  1999/11/26 09:16:28  goldbach
 * Return to skipping the header entry Proxy-Connection. Currently we're unable to
 * process multiple documents per connection (as with "Proxy-Connection:
 * keep-alive"), so we replaced "keep-alive". This didn't work, though.
 * So now we're just skipping the entry.
 *
 * Revision 1.17  1999/11/24 18:15:32  gellert
 * now replacing proxy-connection "keep-alive" with "close"
 * to prevent the transmission of more than one files over
 * one tcp-connection (multiplexing).
 *
 * Revision 1.16  1999/11/11 16:50:32  gellert
 * Now the selective copying of content-length (just in POST-method)
 * works.
 *
 * Revision 1.15  1999/11/11 15:47:13  gellert
 * Now the http header-entries are case insensitive.
 * And the content-length will pass the proxy if the method is
 * "POST" (then the content-length is necessary an the content
 * is not changed by the proxy).
 *
 * Revision 1.14  1999/10/26 14:12:07  gellert
 * just changed a log message
 *
 * Revision 1.13  1999/10/26 14:08:35  gellert
 * now the "Accept-Encoding" field is copied, but not the following
 * value (meaning, that no encoding is accepted).
 *
 * Revision 1.12  1999/10/26 10:59:51  gellert
 * just fixed a bug: the alphabetical order of the header_actions is
 * important (because we use bsearch to search in the struct).
 *
 * Revision 1.11  1999/09/15 09:58:00  goldbach
 * added Mime-Version header. Reformatted X-Cache and Proxy-Connection.
 *
 * Revision 1.10  1999/08/12 13:38:09  gellert
 * added two entries for testing:
 * X-Cache and Proxy-Connection
 *
 * Revision 1.9  1999/08/04 11:43:07  gellert
 * Changed the type check. Now there are two kinds of entries in httpf.cfg:
 * 1) ftype mimetype offset pattern: If the given mimetype matches, than
 *    it is checked whether the block contains pattern at offset.
 * 2) ftype mimetype: If the given mimetype matches, no further check follows.
 *    This makes sense for types like text/plain...
 * These changes required the change of the threadcontext, there is now a
 * character string containing the given mimetype of the server response.
 *
 * Revision 1.8  1999/07/28 14:00:26  gellert
 * if our substitution of tags increases the size of a dokument
 * the browser sents a RESET when the original size (given by
 * Content Length) is reached. So we skip ContentLength in the
 * header.
 *
 * Revision 1.7  1999/07/22 14:27:58  goldbach
 * minor cleanup for logmsg() (again)
 *
 * Revision 1.6  1999/07/21 13:13:40  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.5  1999/06/09 20:33:52  gellert
 * new version. now the parsing of tags and attributes works.
 * and it still works if the client cancels the transmission
 * (stop button in browser).
 *
 * Revision 1.4  1999/05/27 16:47:57  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.3  1999/05/12 14:49:09  goldbach
 * moved find_eol to misc.c
 *
 * Revision 1.2  1999/03/30 12:30:04  goldbach
 * 1st surfable version
 *
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>

#include "httpf.h"
#include "mythreads.h"
#include "analyze_header.h"
#include "misc.h"
#include "log.h"

static int parse_headerline(char *, char**, char**);
static int process_headerline(char *, TCT *);
static int call_funcptr(char *, char *, TCT *);
static int node_compare(const void *, const void *);

/* deal with header lines */
int content_type(char *, TCT *);

/*
 * RFC 2068, 14 Header Field Definitions
 * the alphabetical order is important!!! because we
 * use bsearch to search for the entries...
 */
struct header_action header_actions[] = {
    {"accept", copy_headerline},
    {"accept-charset", copy_headerline},
    {"accept-encoding", copy_fieldname},
    {"accept-language", copy_headerline},
    {"accept-ranges", copy_headerline},
    {"age", copy_headerline},
    {"allow", copy_headerline},
    {"authorization", copy_headerline},
    {"cache-control", copy_headerline},
    {"connection", copy_headerline},
    {"content-base", copy_headerline},
    {"content-encoding", copy_headerline},
    {"content-language", copy_headerline},
    /* if we substitute some tags the document may be
       prolonged. the browser then send a RESET when
       the content length is reached. */
    {"content-length", process_content_length},
    {"content-location", copy_headerline},
    {"content-md5", copy_headerline},
    {"content-range", copy_headerline},
    {"content-type", content_type},
    /* who wants to eat cookies */
    {"cookie", copy_headerline},
    {"date", copy_headerline},
    {"etag", copy_headerline},
    {"expires", copy_headerline},
    {"from", copy_headerline},
    {"host", copy_headerline},
    {"if-match", copy_headerline},
    {"if-modified-since", copy_headerline},
    {"if-none-match", copy_headerline},
    {"if-range", copy_headerline},
    {"if-unmodified-since", copy_headerline},
    {"last-modified", copy_headerline},
    {"location", copy_headerline},
    {"max-forwards", copy_headerline},
    /* not in RFC: */
    {"mime-version", copy_headerline},
    {"pragma", copy_headerline},
    {"proxy-authenticate", copy_headerline},
    {"proxy-authorization", copy_headerline},
    /* not in RFC: */
    /* Proxy-Connection is usualy used instead of the
       Connection header entry. */ 
/*
 * this doesn't work (see the function below), so we'll return
 * to just skipping the whole entry.
     {"proxy-connection", process_proxy_connection}, */

    {"proxy-connection", skip_headerline},
    {"public", copy_headerline},
    {"range", copy_headerline},
    {"referer", copy_headerline},
    {"retry-after", copy_headerline},
    {"server", copy_headerline},
    /* who wants to eat cookies? */
    {"set-cookie", copy_headerline},
    {"transfer-encoding", copy_headerline},
    {"upgrade", copy_headerline},
    {"user-agent", copy_headerline},
    {"vary", copy_headerline},
    {"via", copy_headerline},
    {"warning", copy_headerline},
    {"www-authenticate", copy_headerline},
    /* not in RFC: */
    {"x-cache", copy_headerline}
    
};

/*
 * holds the field name. used when making verbatim copies
 * of header lines. usually the header field function knows
 * what field name it belongs to.
 */

/* static char *field_name; */

/*
 * process_http_header
 *
 * Parse the HTTP header line by line and call the header
 * field function.
 * The header is expected to end with '\r\n'. 
 * Returns 1 on failure, 0 otherwise.
 */
int process_http_header(TCT *tct, int start)
{
    char *line;
    char *nl,  *last_nl;
    char *header_start;
    
    /* let these all point to the start of the header */
    last_nl=nl=header_start=tct->inbuf+start;
    
    /* as long as we didn't reach the end of the buffer contents */
    while((int)(last_nl-tct->inbuf)<tct->inoffset-4) {
	nl = strchr(last_nl, '\n');
	
	/*
	  if ((int)(nl-last_nl)==0) {
	  return 0;
	  }
	*/
	
	/* abort if there's no newline */
	if (nl==NULL) {
	    return 1;
	}
		
	/* point to the start of the current line */
	line = last_nl;
	/* and insert a "0" at the end of line */
	line[nl-last_nl] = 0;
	
	/* null out the \r, too */
	if (nl[-1]=='\r') nl[-1] = 0;
	
	/* call the according function */
	process_headerline(line, tct);
	
	/* skip [\r]\n */
	last_nl=nl+1;	
     }

    /* og, 12.05.1999: and at the end there is a final crlf */
    output_string("\r\n", 2, tct);
    return 0;
}


/*
 * process_headerline
 *
 * Parse a single header line and call the according
 * function with field name, field contents and log level.
 * Returns 1 on failure, 0 otherwise.
 */
static int process_headerline(char *line, TCT *tct)
{

    char *namebuf, *contentbuf;
    int count;
    char *nameptr;

    /* split the line up */
    if (parse_headerline(line, &namebuf, &contentbuf)==1) {
	return 1;
    }

    logmsg(LOG_DEBUG, "Name: %s. Content: %s", namebuf, contentbuf);

    nameptr = &namebuf[0];
    for (count = 0; *nameptr != '\0'; nameptr++) {
          *nameptr = tolower(*nameptr);
          }
                    
    /* call according function */
    if (call_funcptr(namebuf, contentbuf, tct)==1) {
	logmsg(LOG_NOTICE, "[%d] Unknown HTTP header entry: %s", tct->tid, line);
        return 1;
    }

    return 0;
}


/*
 * parse_headerline
 *
 * Fill name and content buffers with the header line data.
 * Returns:
 *   0 -- no error
 *   1 -- separator (colon) not found
 */
static int parse_headerline(char *line, char **nameptr, char **contentptr) {
  size_t contentlen;
  int namelen;
  char *content;

  /* find field name and field contents seperator */
  content=strchr(line, ':');
  if (content==NULL) return 1;

  /* get the lengths */
  /* the lengths have not to be checked because "line" is
     completely in the inbuf and 0-terminated */
  namelen = (content-line);
  /* skip ":" */
  content+=1;
  /* skip a blank after ':' */
  if (content[0]==' ') {
    content+=1;
  }
  contentlen=strlen(content);
    
  /* copy pointer to name and terminate the string in the inbuf */
  *nameptr = line;
  (*nameptr)[namelen]=0;

  /* copy pointer to contents and terminate the string */
  *contentptr = content;
  (*contentptr)[contentlen]=0;
  
  return 0;
}


/*
 * call_funcptr
 *
 * Find matching header_action entry and call the function.
 * Returns 1 on failure, 0 otherwise.
 */
static int call_funcptr(char *name, char *content, TCT *tct)
{
    struct header_action *hap, ha;

    ha.name = name;
    hap = bsearch(&ha, header_actions,
		  sizeof(header_actions)/sizeof(struct header_action),
		  sizeof(struct header_action),
		  node_compare);
    if (hap==NULL) {
	logmsg(LOG_INFO, "[%d] Unable to find matching function for header entry '%s'.", tct->tid, name);
	return 1;
    } else {
	
	/* set global var. used in copy_headerline and
         * emtpy_headerline
         */
        tct->field_name=name;
	
	hap->action(content, tct);
	return 0;
    }
}

/*
 * node_compare
 *
 * Compare name of header_action entries.
 */
static int
node_compare(const void *node1, const void *node2) {
    return (strcmp(
	((const struct header_action *)node1)->name,
	((const struct header_action *)node2)->name));
}

/*
 * copy_headerline
 *
 * Copy the header line as is to the output buffer.
 */
int copy_headerline(char *content, TCT *tct) {
  size_t n; /* chars copied */
    
  logmsg(LOG_DEBUG, "[%d] Copying HTTP header entry '%s: %s'.",
	 tct->tid, tct->field_name, content);
    
  /* field name from thread context */
  n=strlen(tct->field_name);
  output_string(tct->field_name, n, tct);

  /* seperator */
  output_string(": ", 2, tct);

  /* contents */
  n=strlen(content);
  output_string(content, n, tct);

  /* newline */
  output_string("\r\n", 2, tct);

  return 0;
}

/*
 * process_content_length
 *
 * If the Method if POST, then Content-Length has to be provided
 * to the server. So copy it (just in this case).
 */
int process_content_length(char *content, TCT *tct) {
  /* POST method? */
  if (tct->method == 3) {
	logmsg(LOG_DEBUG, "[%d] POST method.",
		   tct->tid, tct->field_name, content);
	copy_headerline(content, tct);
  } else {
	logmsg(LOG_DEBUG, "[%d] No POST method: Skipping 'Content-Length'",
		   tct->tid);
	skip_headerline(content, tct);
  }    
  return 0;
}

/*
 * copy_fieldname
 *
 * Just copy the fieldname, but with an
 * empty content (eg for Accept-Encoding:
 * no content means: no encoding is accepted)
 */
int copy_fieldname(char *content, TCT *tct) {
  size_t n; /* chars copied */
    
  logmsg(LOG_DEBUG, "[%d] Copying HTTP header entry '%s. Content omitted: %s'.",
	 tct->tid, tct->field_name, content);
    
  /* field name from global var */
  n=strlen(tct->field_name);
  output_string(tct->field_name, n, tct);

  /* seperator */
  output_string(": ", 2, tct);

  /* newline */
  output_string("\r\n", 2, tct);

  return 0;
}

/*
 * process_proxy_connection
 *
 * the value "keep-alive" has to be replaced by "close".
 * otherwise the tcp connection is used by more than one
 * http request (multiplexing). We don't want to keep
 * track of that...
 *
 * Update: unfortunately this doesn't work. Maybe we'll need
 * this function in the future (read: use it as a template
 * to replace header values)...
 * It is unused.
 */
int process_proxy_connection(char *content, TCT *tct) {
  size_t n; /* chars copied */
  char *contentptr;
  int count;
    
  logmsg(LOG_DEBUG, "[%d] Processing HTTP header entry '%s'.",
	 tct->tid, tct->field_name, content);
    
  /* get field name */
  n=strlen(tct->field_name);
  output_string(tct->field_name, n, tct);

  output_string(": ", 2, tct); /* seperator */
  n=strlen(content); /* no contents */

  contentptr = &content[0];
  for (count = 0; *contentptr != '\0'; contentptr++) {  
    *contentptr = tolower(*contentptr); }

  if (strncmp(content, "keep-alive", 10)==0) {
    logmsg(LOG_DEBUG, "[%d] Proxy-Connection: keep-alive replaced.",
	   tct->tid);
    output_string("close", 5, tct);
  } else {
    output_string(content, n, tct);
  }

  output_string("\r\n", 2, tct);  /* newline */
  
  return 0;
}

/*
 * skip_headerline
 *
 * Don't send this header entry to the output buffer.
 */

int skip_headerline(char *content, TCT *tct)
{
    return 0;
}


/*
 * content_type
 *
 * Copy the header line as is to the output buffer.
 * Print content type.
 * Set mime type id in thread context.
 */
int content_type(char *content, TCT *tct) {
  copy_headerline(content,tct);
  logmsg(LOG_INFO, "[%d] Content-Type is '%s'.", tct->tid, content);

  strncpy(tct->mime_type, content, min(strlen(content)+1, LEN_MIMETYPE-1));
  tct->mime_type[LEN_MIMETYPE] = '\0';
  logmsg(LOG_INFO, "[%d] Mime_Type: %s", tct->tid, tct->mime_type);
  if (strcmp(content, "text/html")==0) {
      logmsg(LOG_INFO, "[%d] Receiving HTML data.", tct->tid);
  } else {
      logmsg(LOG_INFO, "[%d] Receiving non-HTML data.", tct->tid);
  }
  
  return 0;
}

