/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */
/*
 *
 * $Id: analyze_request_header.c,v 1.9 2001/03/01 15:51:36 glauschwuffel Exp $
 *
 * $Log: analyze_request_header.c,v $
 * Revision 1.9  2001/03/01 15:51:36  glauschwuffel
 * After simplifying read.c we introduced a bug in
 * analyze_request_header: it always said that the client didn't send any
 * body.
 *
 * Revision 1.8  2000/06/19 21:38:56  glauschwuffel
 * includes log.h
 *
 * Revision 1.7  2000/03/07 15:06:46  goldbach
 * added GPL info
 *
 * Revision 1.6  2000/02/21 14:35:38  gellert
 * Just a change in the loglevel of some messages.
 *
 * Revision 1.5  1999/07/22 14:27:59  goldbach
 * minor cleanup for logmsg() (again)
 *
 * Revision 1.4  1999/07/21 13:19:39  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.3  1999/05/27 16:47:58  goldbach
 * minor cleanup for logmsg()
 *
 * Revision 1.2  1999/05/12 16:04:19  goldbach
 * major clean-up session
 *
 * Revision 1.1  1999/03/30 12:30:05  goldbach
 * 1st surfable version
 *
 *
 */

#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <unistd.h>
#include <string.h>

#include "httpf.h"
#include "mythreads.h"
#include "analyze_header.h"
#include "analyze_request_header.h"
#include "misc.h"
#include "log.h"

/*
 * the methods as defined in RFC 2068
 */
static char methods[][7]={
    "GET", "PUT", "POST", "HEAD", "DELETE",
    "TRACE", "LINK", "UNLINK"};

/*
 * analyze_request_header()
 *
 * Analyzes the HTTP header information line by line.
 *
 * Returns 0 if there's no body, 1 otherwise.
 */

int analyze_request_header(TCT *tct)
{
    int header_start;
    
    /* process request line and get start of optional header */
    header_start = process_http_request_line(tct);
    if (header_start==0)
	return 1;

    logmsg(LOG_DEBUG, "[%d] Optional HTTP header starts at offset %d.",
	   tct->tid, header_start);
    
    /* FIXME: fiddle around with the request line */

    /* send request line to output buffer */
    write_request_line(tct);

    /* end of input buffer reached? */
    if (tct->inlen==header_start)
	return 0;
    
    /* proccess header. '(void)' because we don't care if there are errors CHECK */
    (void)process_http_header(tct, header_start);

    /* check if we have a body */
    if (tct->inoffset==tct->inlen)
      return 0; /* no body */
    else
      return 1; /* body */
}



/*
 * process_http_request_line
 * 
 * Examine the first header line and look for the method,
 * version and the URI.
 * Returns:
 *   0 -- no method found
 *   >0 -- start of optional header
 * Note: Knows methods as listed in RFC 2068
 */

int process_http_request_line(TCT *tct)
{
    int id;
    char *uri,*v;
    char *eol; /* end of the request line */

    
    /* find next line and replace CRLF with null chars */
    eol=find_eol(tct->inbuf, HTTP_BUFSIZE, 1);
    eol++; /* go to first char of next line */

    /* find the SP after the method (before the uri) */
    uri=strchr(tct->inbuf, ' ');
    /* 
     * no method? leave!
     * NB: we modified the inbuf...
     */
    if (uri==0) {
	logmsg(LOG_ERR, "[%d] Request header has no method.", tct->tid);
	return 0;
    }
    
    /* move from space position to first char of uri */
    uri++;
    tct->uri = uri;
    
    /* split method and the rest*/
    uri[-1]=0;
    
    /*
     * don't use bsearch or something like that,
     * this could be faster... most common methods
     * are at the beginning
     */
    for (id=0; id<=7; id++) 
	if (strncmp(tct->inbuf, methods[id],strlen(tct->inbuf))==0)
	    break;

    /* store method id */
    tct->method=id+1;
    
    /* retrieve http version */
    v=strrchr(uri, ' ');
    if (v==0) {
	logmsg(LOG_ERR, "[%d] HTTP version missing, assuming builtin version.", tct->tid);
	tct->version_major=BUILTIN_HTTP_MAJOR;
	tct->version_minor=BUILTIN_HTTP_MINOR;
    } else {
	v++; /* go to first char of version string */
	v[-1]=0; /* split uri and version */
	
	/* get version information. if this fails, assume builtin version */
	if(sscanf(v, "HTTP/%d.%d", &(tct->version_major), &(tct->version_minor))!=2) {
	    logmsg(LOG_ERR, "[%d] Broken HTTP version (%s), assuming builtin version.",
		   tct->tid, v);
	}
    }
    
    logmsg(LOG_INFO,"[%d] Received client request: method %s, version %d.%d, URI %s.",
	   tct->tid, methods[tct->method-1], tct->version_major, tct->version_minor, tct->uri);

    /* optional header starts here */
    return (int)(eol-tct->inbuf);
}

/*
 * write_request_line
 *
 * write method, uri and version from context to outbuf
 */
void write_request_line(TCT *tct)
{
    char version[20];
    
    sprintf(version, " HTTP/%d.%d", tct->version_major,
	    tct->version_minor);
    
    output_string(methods[tct->method-1], strlen(methods[tct->method-1]),tct);
    output_string(" ",  1, tct);
    output_string(tct->uri, strlen(tct->uri), tct);
    output_string(version, strlen(version), tct);
    output_string("\r\n", 2, tct);
    logmsg(LOG_ERR, "Request: %s %s %s", methods[tct->method-1], tct->uri, version);
}

