/*
 * $Id: args.c,v 1.1 2001/03/01 16:06:29 glauschwuffel Exp $
 *
 * args.c -- get command line arguments
 *
 * $Log: args.c,v $
 * Revision 1.1  2001/03/01 16:06:29  glauschwuffel
 * argument processing module. added option 'f' (config file) and 'V'
 * (logging facility
 *
 */

#include <sys/socket.h>
#include <arpa/inet.h>
#include <sys/param.h>
#include <sys/resource.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <netdb.h>

#include "mythreads.h" /* YES, NO */
#include "log.h"
#include "args.h"

ARGS args;

extern int log_level; /* from log.c */
extern int resource_flag ; /* flag for child: run setrlimit after forking */

struct rlimit maxfiles; /* set_resource_limit() */

static int facility_table[] = {
  LOG_LOCAL0,
  LOG_LOCAL1,
  LOG_LOCAL2,
  LOG_LOCAL3,
  LOG_LOCAL4,
  LOG_LOCAL5,
  LOG_LOCAL6,
  LOG_LOCAL7,
};


void set_default_args() {
  args.number_of_threads          = DEFAULT_NUMBER_OF_THREADS;
  args.server_socket_size         = DEFAULT_SOCKET_SIZE;
  args.client_socket_size         = DEFAULT_SOCKET_SIZE;
  args.copy_size                  = DEFAULT_COPY_SIZE;
  args.daemon_flag                = YES;
  args.server_port                = DEFAULT_SERVER_PORT;
  log_level = args.log_level      = DEFAULT_LOG_LEVEL;
  args.real_proxy_flag            = DEFAULT_REAL_PROXY_FLAG;
  args.deadlock_checker_intervall = DEFAULT_DEADLOCK_CHECKER_INTERVALL;
  args.recv_duration              = DEFAULT_RECV_DURATION;
  args.log_facility               = DEFAULT_LOG_FACILITY;

  strncpy(args.config_file, DEFAULT_CONFIG_FILE, MAX_CONFIG_FILE_NAME_LENGTH);
}

void usage() {
  printf("Usage:\n"
	 "\thttpf   [-h] [-?] [-p <port>] [-m <copysize>] [-t <threads>]\n"
	 "\t\t[-s <clientside socket size>] [-S <serverside socket size>]\n"
	 "\t\t[-d <destination host> -D <destination port>]\n\n"
	 "Options (default)\n"
	 "\t-h | -?                   Displays this message\n"
	 "\t-v <verbosity>      (5)   Set log level level (0-7)\n"
	 "\t-V <facility>       (0)   Set log facility to LOG_LOCALn (0<=n<=7)"
         "\t-p <port>        (1080)   Listen on specified port\n"
         "\t-m <copysize>    (8192)   Maximum amount of data [byte] copied by proxy\n"
         "\t                          during each `read-write'-cycle\n"
	 "\t-n                        No daemon: stay in foreground.\n"
	 "\t-t <threads>       (50)   number of pre-forked threads\n"
	 "\t-s <socket size>    (-)   Socket size [byte] for connection with client\n"
	 "\t-S <socket size>    (-)   Socket size [byte] for connection with server\n"
	 "\t                          System specific default socket sizes may vary\n"
	 "\t-d <host>           (-)   Destination host for `Plug-GW' mode\n"
	 "\t-D <port>           (-)   Port to connect on destination host in `Plug-GW' mode\n"
	 "\t-f <config> (httpf.cfg)   Configuration file\n"
	 );
}

void process_args(int argc, char **argv) {
  int c;

  while ((c=getopt(argc, argv, 
		   ARGUMENT_STRING
		   )) != EOF) {
    switch (c) {
    case '?':
    case 'h':
    default:
      usage();
      exit(1);
    case 'v': /* change level of verbosity */
      log_level = args.log_level = atoi(optarg);
      break;
    case 'V': { /* change log facility */
      int n=atoi(optarg);

      if ((n<0) || (n>7)) {
	logmsg(LOG_ERR,"log facility out of bounds: must be between 0 and 7");
	exit(EXIT_FAILURE);
      } else {
	args.log_facility = facility_table[n];
      }
      
      break;
    }
    
    case 'p': /* Port an dem der Proxy "listen" soll */
      args.server_port=atoi(optarg);
      break;
    case 'm': /* Read/Write-Laenge beim Kopieren */
      args.copy_size = atoi(optarg);
      break;
    case 'n': /* No Daemon, stay in foreground... */
      args.daemon_flag = NO;
      break;
    case 's': /* Client-Side Socket Groesse */
      args.client_socket_size = atoi(optarg);
      break;
    case 'S': /* Server-Side Socket Groesse */
      args.server_socket_size = atoi(optarg);
      break;
    case 'f':
      strncpy(args.config_file, optarg, MAX_CONFIG_FILE_NAME_LENGTH);
      break;
    case 't': /* Anzahl der Threads */
      args.number_of_threads = atoi(optarg);

      if (args.number_of_threads >= 0) {
#define FO 10 /* config, libs, 2 im Sinn ;) */
	int files_needed;
	
	getrlimit (RLIMIT_NOFILE, &maxfiles);
	files_needed = (args.number_of_threads*2) + FO;
	if (files_needed > maxfiles.rlim_max) {
	  logmsg(LOG_WARNING, "Number of threads exceeds available file descriptors");
	  logmsg(LOG_WARNING, "Limiting threads to %lu", maxfiles.rlim_max - FO);
	  args.number_of_threads = (maxfiles.rlim_max - FO);
	  maxfiles.rlim_cur = maxfiles.rlim_max;
	  resource_flag = 1;
	} else if (files_needed > maxfiles.rlim_cur) {
	    maxfiles.rlim_cur = files_needed;
	    resource_flag = 1;
	}
	/* An dieser Stelle soll eigentlich der Bedarf an Filehandles
	 * fuer diesen Prozess und alle seine Kindprozesse gesetzt 
	 * werden (pro Thread brauch man mindestens einen Socket, 
	 * zuzueglich ein paar Handles fuer die Standardfiles und
	 * fuer das Logfile.
	 *
	 * Leider setzt der Aufruf von 'Set Session ID' (setsid()) 
	 * im Hauptprogramm unter Solaris 2.5.1 diese Einstellungen 
	 * zurueck. Mann muss diesen Aufruef (setrlimit()) also nach
	 * setsid() wiederholen.
	 * 
	 * Da ich nicht weiss, ob dies ein Bug oder ein Feature von
	 * Solaris ist, lasse ich die Einstellungen trotzdem hier, da
	 * dies die geeignete Stelle zu sein scheint.
	 */
	set_resource_limit();
      } else return;
      break;
      
      break;
    case 'd': /* Zielhost Adresse, wenn Verbindungen immer nur zu 
	       * einem Zielrechner erlaubt werden (wie plug-gw) 
	       */
      args.real_proxy_flag = 0;
      strncpy(args.destination_host, optarg, sizeof(args.destination_host) );
      if (((args.destination_address).sin_addr.s_addr = inet_addr(args.destination_host)) == -1) {
	/* vielleicht ein DNS-Name? */
	struct hostent		*hp;
	if ((hp = gethostbyname(args.destination_host)) == NULL) {
	  logmsg(LOG_ERR, "Can't resolve destination host.");
	  exit(1);
	} else {
	  /* DNS Adresse */
	  bcopy(hp->h_addr,
		(char *)&(args.destination_address).sin_addr,
		hp->h_length);
	  (args.destination_address).sin_family = hp->h_addrtype;
	}
      } else {
	/* IP Adresse */
	(args.destination_address).sin_family = AF_INET;
      }
      break;
    case 'D': /* Port an den auf dem Zielhost verbunden werden soll */
      (args.destination_address).sin_port = htons(atoi(optarg));
      break;
    }
  }
}

void set_resource_limit() {
  if (setrlimit (RLIMIT_NOFILE, &maxfiles)) {
    logmsg(LOG_ERR, "Unable to adjust the number of file descriptors.");
    logmsg(LOG_ERR, "setrlimit(RLIMIT_NOFILE, ...)");
    exit (EXIT_FAILURE);
  }
}
