/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Log: log.c,v $
 * Revision 1.3  2001/03/01 15:58:29  glauschwuffel
 * logging facility is now defined by args.c
 *
 * Revision 1.2  2000/09/06 13:19:23  glauschwuffel
 * syslog is protected by the semaphore \(again\). Now the syslogd is happy again and does *not* run 5 seconds for every call to logmsg
 *
 * Revision 1.1  2000/06/19 22:06:33  glauschwuffel
 * Major overhaul of the HTML parser. Now we have something like a
 * state machine (yes, I know that I should've done it that way
 * in the first place ;)).
 *
 * Split up some modules (misc.c, copy_pkts.c). filter_html.c
 * is much more cleaner now. I think that everyone should be able
 * to understand how the HTML parser works now. Especially
 * the skipping of comments is easy to get now ;)
 *
 * This seems to be getting cleaner...
 *
 *
 * $Id: log.c,v 1.3 2001/03/01 15:58:29 glauschwuffel Exp $
 */

#include "log.h"
#include "args.h"

sem_t log_sema;
int log_level=DEFAULT_LOG_LEVEL;

extern ARGS args;


/*
 * initialize_logging()
 *
 * Initialize the semaphore and open the log file.
 */
void initialize_logging(void) {

  /* semaphore init */
  sem_init(&log_sema, 0, 1);
  openlog("httpf", LOG_PID, args.log_facility);
}

/*
 * logmsg
 *
 * Log information (via syslog()) if given logging level is
 * not higher than the global level (LOG_DEBUG is high).
 * Uses vsnprintf() and a semaphore.
 *
 * Inputs:
 *   log_level -- passed to syslog().
 *   fmt -- printf-like format string.
 *   ... -- args needed by format string
 */

int logmsg (int level, const char *fmt, ...) {
    va_list ap;
    char logbuf[SIZE_LOGBUF];
    int result;
    
    /* don't log if level is too high */
    if (level > log_level) return 0;

    va_start(ap, fmt);

    sem_wait(&log_sema); /* wait for semaphore */
    /* format string */
    result = vsnprintf(logbuf, SIZE_LOGBUF, fmt, ap);
    /* the %s prevents the content of logbuf to be interpreted */
    syslog(level, "%s", logbuf);

/*
 * protect syslog with the semophore, too.
 * syslog() itself is MT-safe, however, we access
 * a local buffer that's shared between the threads.
 * Don't protect syslog with the semaphore here and
 * you'll notice a very  high running time.
 */
    sem_post(&log_sema);
    
    if (result>SIZE_LOGBUF) {
      syslog(level, " log_msg: message was truncated");
    }
      
    va_end(ap);

    return 0;
}

