/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Log: parse_attributes.c,v $
 * Revision 1.7  2001/03/01 16:00:56  glauschwuffel
 *  check_protocol now checks 'action', too. So we now have 'href', 'src' and 'acti
 *  Maybe we should make this a positive list, too?
 *
 * Revision 1.6  2000/09/15 15:04:48  glauschwuffel
 * find_attrstart now sets the inoffset attribute. If the first char
 * of a new attribute is an illegal character we don't enter an infinite
 * loop anymore...
 *
 * Revision 1.5  2000/09/15 15:00:53  glauschwuffel
 * comments
 *
 * Revision 1.4  2000/09/06 13:22:08  glauschwuffel
 * when do_attr returns with a failure and we NEED_MORE_INPUT we return immediately. Otherwise quoted attribute values would lead to an infinite loop for this thread
 *
 * Revision 1.3  2000/06/28 22:01:09  glauschwuffel
 * parse errors are processed smoothly ;)
 *
 * Revision 1.2  2000/06/21 18:48:44  glauschwuffel
 * find_valend now skips to next WS or '>' after quoted value.
 * This ensures that
 * <p onMouseOver="men(1)"; onMouseOut="men(0)";>
 * will be parsed.
 * Adapted call to proto_allowed and simplified the call of
 * this function.
 *
 * Revision 1.1  2000/06/19 22:06:33  glauschwuffel
 * Major overhaul of the HTML parser. Now we have something like a
 * state machine (yes, I know that I should've done it that way
 * in the first place ;)).
 *
 * Split up some modules (misc.c, copy_pkts.c). filter_html.c
 * is much more cleaner now. I think that everyone should be able
 * to understand how the HTML parser works now. Especially
 * the skipping of comments is easy to get now ;)
 *
 * This seems to be getting cleaner...
 *
 *
 * $Id: parse_attributes.c,v 1.7 2001/03/01 16:00:56 glauschwuffel Exp $
 */

#include <string.h>
#include "misc.h"
#include "mythreads.h"
#include "taghash.h"
#include "cfglists.h"
#include "parse_tools.h"
#include "log.h"

static int do_attr(TCT*);
static int process_attr(char*,char*,TCT*);

static char *find_attrstart(char*,TCT*);
static char *find_attrend(char*,TCT*);
static char *find_valstart(char*,TCT*);
static char *find_valend(char*,TCT*); 
static int check_for_script(char*);
static int check_protocol(char*,char*,TCT*);
static char *jump_to_end_of_attribute(char*,TCT*);

/*
 * parse_attributes()
 *
 * Parse the tag's attributes.
 * Returns FAILURE on error, check error attribute.
 */
int parse_attributes(TCT*tct) {
  char *ws;

  logmsg(LOG_DEBUG, "[%d] Processing attributes.", tct->tid);

  while(tct->error==NO_ERROR) {
    if (do_attr(tct)==FAILURE) {

      logmsg(LOG_NOTICE,
	     "[%d] The last attribute was parsed with error %d",
	     tct->tid,tct->error);

      /* return immediately if we need more data */
      if (tct->error==NEED_MORE_INPUT)	return FAILURE;

      /*
       * bad chars, skip to next whitespace.
       * if we reach the end of the buffer we
       * need more input and leave
       */
      ws=jump_to_end_of_attribute(tct->inbuf+tct->inoffset,tct);
      if(tct->error!=NO_ERROR) {
 	tct->error=NEED_MORE_INPUT;
	return FAILURE;
      }
      
      /* set continuation point */
      tct->inoffset=ws-tct->inbuf;
    } else {
      /* attribute processed. Are we at the end of an attribute? */
      /* skip possible WS between attr and closing '>' */
      if (legal_whitespace(tct->inbuf[tct->inoffset])) {
	ws=skip_leading_whitespace(tct->inbuf+tct->inoffset,tct);
	if(tct->error!=NO_ERROR) {
	  tct->error=NEED_MORE_INPUT;
	  return FAILURE;
	}
	/* copy the WS. remove this if the resulting
           document ought to look cleaner.
	   if you do this you have to insert a
           output_string(" ", 1, tct) instead or the
           attributes won't be seperated ;) */
	output_string(tct->inbuf+tct->inoffset,
		      ws-(tct->inbuf+tct->inoffset),tct);
	tct->inoffset=ws-tct->inbuf;
      }

    }

    /* end of tag? */
    if (tct->inbuf[tct->inoffset]=='>') break;

  } /* while */

  tct->inoffset++; /* skip the closing '>' */
  output_string(">", 1, tct); /* close the tag */
  /* we're leaving the tag and attribute parser */       
  tct->parse_state = PARSE_CONTENT;

  return OKAY;
}



/*
 * do_attr
 *
 * Try to get and handle the next attribute in a string.
 * Returns:
 *  0  -- abort parsing this string. 
 *        The attr is not allowed so skip the rest.
 *  FAILURE -- there was an error (illegal characters in attribute or value).
 *  other --  the string right behind the attribute's value or the string
 *            right behind the attribute (if there was no value)
 *
 * Note: we modify the input buffer. The attribute and value string
 *       are terminated with a null character.
 */
static int do_attr(TCT *tct) {
  char *attr_start, *attr_end, *val_start=NULL, *val_end;

  /* find start of attribute */
  attr_start = find_attrstart(tct->inbuf+tct->inoffset,tct);
  if(attr_start==NULL) return FAILURE;

  /* find end of attribute */
  attr_end = find_attrend(attr_start,tct);
  if (attr_end==NULL) return FAILURE;

  /* try to find value if we are not at the end of a tag */
  if (attr_end[0]!='>') {
    val_start=find_valstart(attr_end,tct);
    if (tct->error!=NO_ERROR) return FAILURE;
  }

  if (val_start==NULL) { /* no value */
    char a;

    /* store char and terminate attribute */
    a=attr_end[0]; attr_end[0]=0;
    if (tct->tag_action!=TAG_SKIP)
      if (process_attr(attr_start, NULL, tct)==FAILURE) {
	logmsg(LOG_ERR, "[%d] This is a bug: FIXME", tct->tid);
      };

    attr_end[0]=a; /* put char back */
    tct->inoffset=attr_end-tct->inbuf;
  } else { /* we have a value */
    char a,v;

    val_end=find_valend(val_start, tct);
    if (val_end==NULL) return FAILURE;

    /* terminate attribute and value */
    v=val_end[0]; a=attr_end[0];
    attr_end[0]=0; val_end[0]=0;

    if (tct->tag_action!=TAG_SKIP)
      if (process_attr(attr_start, val_start,tct)==FAILURE) {
	logmsg(LOG_ERR, "[%d] This is a bug: FIXME", tct->tid);
      };

    val_end[0]=v; attr_end[0]=a; /* put them back */
    tct->inoffset=val_end-tct->inbuf;
  }
  return OKAY;
}


/*
 * process_attr
 *
 * Perform configured action for this attribute.
 * 'replabort' sets tag_action to TAG_SKIP,
 * the other actions set it to TAG_PROCESS.
 */
static int process_attr(char*attrbuf, char *valbuf, TCT *tct) {
  char *replacement; /* from hash-table */
  int action; /* from hash-table */
    
  /* find out what we are supposed to do */
  action = attrib_action(attrbuf);
    
  if (action < 0) {
    /* action not specified in configuration file */
    logmsg(LOG_INFO, "[%d] Not allowing attribute '%s'.", tct->tid, attrbuf);
    return OKAY;
  }

  /* check if we have some script code in the value */
  if (check_for_script(valbuf)!=0){
    logmsg(LOG_INFO, "Found script code in value of attribute %s:", attrbuf);
    logmsg(LOG_INFO, "\"%s\"", valbuf);
    return OKAY;
  }

  /* logging? (1) */
  {
    int level;
    if (action & ACT_LOG) level=LOG_NOTICE; else level=LOG_DEBUG;
    
    if (valbuf!=NULL) {
      logmsg(level, "[%d] Processing attr '%s' with value '%s'",
	     tct->tid, attrbuf, valbuf);
    } else
      logmsg(level, "[%d] Processing attr '%s'",
	     tct->tid, attrbuf);
  }
  
  if (action & (ACT_REPLABORT | ACT_REPLCONT)) {
    /* get replacement for this attr*/
    replacement=attrib_replacement(attrbuf);
    if (action & ACT_LOG) {
      logmsg(LOG_NOTICE, "[%d] Replacing with '%s'.",
	     tct->tid, replacement);
    }
	  
    output_string(replacement, strlen(replacement),tct);

    /* set SKIP action if we should replace and abort */
    if (action & ACT_REPLABORT) tct->tag_action=TAG_SKIP;
    return OKAY;
  } else if (action & ACT_ALLOW) { /* allow attribute */
    /* first check if there's a protocol */
    if (check_protocol(attrbuf, valbuf,tct)==FAILURE) {
      logmsg(LOG_NOTICE, "[%d] Suppressing attribute '%s'", tct->tid, attrbuf);
    } else {
      if (action & ACT_LOG) {
	logmsg(LOG_NOTICE, "[%d] Allowing attribute.", tct->tid);
      }
	  
      /* copy attr 'as is' to output buffer */
      output_string(attrbuf, strlen(attrbuf), tct);
      if (valbuf!=NULL) {
	/* we have a value */
	output_string("=", 1, tct);
	output_string(valbuf, strlen(valbuf), tct);
      }
    }
    tct->tag_action=TAG_PROCESS;
    return OKAY; /* proceed with the attributes */
  } else if (action & ACT_LOG) {
    /* logging has been done at (1) */
    tct->tag_action=TAG_PROCESS;
    return OKAY; /* proceed with the attributes */
  }
    
  logmsg(LOG_ERR, "[%d] Unknown attribute action %d.", tct->tid, action);
  return FAILURE;
}

/*
 * check_for_script
 *
 * Checks the value buffer for embedded script code.
 * (This script code is introduced by "&{".)
 * Returns 0 if the buffer does not contain "&{",
 * -1 otherwise.
 * Notes: Safe to call with a value of NULL for the buffer.
 */
static int check_for_script(char *valbuf) {
  char *pos;

  /* return ok if we have no value */
  if (valbuf==NULL)
    return 0;
  
  pos = valbuf;

  /* look for ampersand */
  while ((pos = memchr(pos, '&', strlen(pos))) != NULL) {
    if (strlen(pos)>1) {
      if (pos[1]=='{') return -1;	
    }
    pos++;
  }

  return 0; /* no script code found */
}


/*
 * find_attrstart
 *
 * Find start of attribute.
 * Takes pointer to buffer that hopefully contains attributes.
 * Leading whitespace is skipped.
 * Returns pointer to first character of attribute or NULL if
 * attribute starts with illegal characters.
 */
static char *find_attrstart(char *buf, TCT *tct) {
  char *attr_start;

  /* skip whitespace */
  attr_start = skip_leading_whitespace(buf,tct);
  if (tct->error!=NO_ERROR) return NULL;
  /*
   * Point to the start of the new attribute.
   * This is important: if we don't set it we'll end up 
   * in an infinite loop.
   */
  tct->inoffset = attr_start-tct->inbuf;
  
  /*
   * Check if the first character of the attribute
   * is legal. Set the recovery point if this is the case,
   * Return with an error if this is not the case.
   */
  if(legal_constituent(attr_start[0])) {
    /* set recovery point */
    tct->last_valid_attr = attr_start;
    return attr_start;
  } else {
    tct->error = UNALLOWED_CHAR;
    return NULL;
  }
}

/*
 * find_attrend
 *
 * Find end of attribute.
 * Takes pointer to beginning of attribute. 
 * Result points right behind last character of attribute or NULL if
 * attribute contains illegal characters.
 */
static char *find_attrend(char *attr_start,TCT*tct) {
  int i=0;

  if (check_inbuf_pointer(attr_start,tct)!=NO_ERROR) return NULL;

  /* Find end of legal characters. */
  while(legal_constituent(attr_start[i])) {
    i+=1;
  }

  if (check_inbuf_pointer(attr_start+i,tct)!=NO_ERROR) return NULL;

  /*
   * an attribute name ends with
   *  - '='
   *      we have a value
   *  - whitespace (there might be a value)
   * - '>'
   *     This is the last attribute
   */
  if (attr_start[i]=='=' ||
      attr_start[i]=='>' ||
      legal_whitespace(attr_start[i]))
    return attr_start+i;
  else {
    tct->error=UNALLOWED_CHAR;
    return NULL;
  }
}

/*
 * find_valstart
 *
 * Find start of (optional) value for an attribute.
 * BUF points to the start of the current attribute.
 * Set the error attribute to UNALLOWED_CHAR if
 * value starts with illegal character.
 *
 * Notes: we allow a hash ('#') at the start of a value
 *   if it is not followed by an ampersand ('&') since this
 *   would allow entities...
 *
 * Returns:
 *  0 - no value
 *  other - start of value.
 */

static char *find_valstart(char *attr_end, TCT *tct) {
  char *cont;
  
  if (check_inbuf_pointer(attr_end,tct)!=NO_ERROR) return NULL;
  
  /* we should have a '=' or a whitespace now */
  if (legal_whitespace(attr_end[0]) != 0) {
    cont=skip_leading_whitespace(attr_end,tct);
    if (tct->error!=NO_ERROR) return NULL;
  } else {
    cont=attr_end;
  }

  /* no value if we are at end of string */
  if (cont[0]=='>') return NULL;

  /*
   * now the only valid continuations are:
   *   '=' -- we have a value
   *   legal_constituent -- new attribute
   * Other values are an error.
   */
  if (cont[0]=='=') {
    /* we have a value */
    cont = skip_leading_whitespace(cont+1,tct);
    if (tct->error!=NO_ERROR) return NULL;
    if (legal_constituent(cont[0]) != 0 ||
	(cont[0]=='#' && cont[1]!='&') || /* no entities ;) */
	a_quote(cont)) {
      /* valstart is legal, set recovery point */
      tct->last_valid_value = cont;
      return cont;
    } else {
      /* valstart is illegal */
      tct->error=UNALLOWED_CHAR;
      return NULL;
    }
  }
  
  /* we have no value or a malformed input */
  if (legal_constituent(cont[0]) != 0)
    return NULL; /* no value (new attr) */
  else
    tct->error=UNALLOWED_CHAR;
    return NULL;
}

/*
 * find_valend
 *
 * Find end of value of current attribute.
 * Inputs:
 *  valstart - result of find_valstart
 * Returns:
 *  address of end of value or NULL on error.
 */
static char *find_valend(char *valstart, TCT *tct) {
  char c;

  if (check_inbuf_pointer(valstart,tct)!=NO_ERROR) return NULL;

  /*
   * First we check for a heading quote.
   * These comparisons are based on the HTML 4.0
   * specs (3.2.2 Attributes, p. 25).
   * This function assumes this:
   *   - quotes should be nested correctly
   *   - &#39 is treated as a single quote and
   *     &#34 and &quot; as double quote. Thus: the string
   *       "bla&quot;
   *     looks awkward, however, it is correct.
   */
  c = a_quote(valstart);
  if (c) { /* we have a quote */
    char *s=find_end(valstart+1, c, tct);
    if (s==NULL) return NULL; /* need more input */

    /*
     * We have a pointer to the end
     * of the quoted value by calling
     * find_end(). We extend the
     * value to anything up to the closing
     * '>' or whitespace. This ensures that
     * something like
     * <p onMouseOver="men(1)"; onMouseOut="men(0)";>
     * will be parsed.
     */
    /* return new pointer to WS or '>' */
    return jump_to_end_of_attribute(s,tct);
  } else {
    /* no quote */
    return find_unquoted_end(valstart+1,tct);
  }
}


/*
 * check_protocol
 *
 * Checks if a protocol is allowed or not. It is allowed if it's
 * a reference without protocol or a reference with a
 * configured protocol.
 * Note:
 *   Currently this works only with the colon character ':' as protocol
 *   delimiter. We search &#xx, too. At least &#58 (that's ":")
 *   If the protocol name contains any &#xx, proto_allowed will not
 *   find them (and thus not allow them)
 */

static int check_protocol(char *attr, char *value,TCT*tct) {
  char *temp, *temp1, *temp2;

  /* do we have a value? */
  if (value == NULL) return OKAY;

  /* do we have a link, a src or an associated action? */
  if ((strncasecmp("href", attr, 5) != 0) &&
      (strncasecmp("src", attr, 4)  != 0) &&
      (strncasecmp("action", attr, 6)  != 0)) {
    return OKAY;
  }
    
  /* look for a colon in the attribute's value */
  temp1 = memchr(value, ':', strlen(value));
  temp2 = strstr(value, "&#58");

  if ((temp1 == NULL) && (temp2 == NULL)) {
    /* there is no colon */
    /* we don't have a protocol, so the reference is okay */
    return OKAY;
  }

  temp = value;
  while ((*temp== ' ') || (*temp== '\"') || (*temp=='\'')) {
    temp++;
  }
    
  return proto_allowed(temp,tct);
}

/*
 * jump_to_end_of_attribute()
 *
 * Given a pointer within the inbuf we jump to
 * the next whitespace or '>' character.
 * Note that this function doesn't care if it's in a
 * quoted attribute or not...
 *
 * Since we step through the buffer one character at a time
 * the only error 'check_inbuf_pointer' will find is 'end of buffer'.
 */
static char *jump_to_end_of_attribute(char*s, TCT*tct) {
  int i=0;
  if(check_inbuf_pointer(s,tct)!=NO_ERROR) return NULL;
  while (s[i] && s[i]!='>' && !legal_whitespace(s[i])) i++;
  if(check_inbuf_pointer(s+i,tct)!=NO_ERROR)
    return NULL;
  else
    return s+i;
}

