/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: parse_tools.c,v 1.9 2001/05/04 12:12:47 gellert Exp $
 *
 * $Log: parse_tools.c,v $
 * Revision 1.9  2001/05/04 12:12:47  gellert
 * Bug-Fix in the binary-check! If there was a no-printable (binary) byte,
 * the procedure returned with an error without clearing the semaphore!
 *
 * Revision 1.8  2001/03/03 17:25:51  glauschwuffel
 * The check for binary data declared as text/html is now available on
 * Linux, too (protected unsafe rand() with semaphore).
 *
 * Expanded the error page when binary data is found a bit:
 * the contents surrounding the suspicious byte is now displayed.
 *
 * Revision 1.7  2001/03/01 16:01:43  glauschwuffel
 * The inbuf_contains_binary predicate now doesn't check the HTTP header
 * ;) This was due to the fact that we didn't keep in mind that the data
 * began at offset inoffset...
 *
 * Revision 1.6  2000/09/15 15:01:57  glauschwuffel
 * comments
 *
 * Revision 1.5  2000/09/06 13:17:27  glauschwuffel
 * find_end did not set the error attribute. Quoted values for attributes would make the thread loop infinitely...
 *
 * Revision 1.4  2000/06/28 22:02:11  glauschwuffel
 * comments containing an odd number of hyphens do not break the proxy anymore
 *
 * Revision 1.3  2000/06/22 09:32:00  glauschwuffel
 * removed tct...count again. init does not belong here...
 *
 * Revision 1.2  2000/06/21 18:50:07  glauschwuffel
 * Adjusted log messages.
 * Changed skip_to_end_of_tag() to meet The Real World:
 * break if there's a '<' that's not in a string.
 * Again this is a Real World effect from Slashdot:
 * <BR><BR<P>content
 * Yes, otherwise the document will be skipped to the end ;(
 *
 * Revision 1.1  2000/06/19 22:06:33  glauschwuffel
 * Major overhaul of the HTML parser. Now we have something like a
 * state machine (yes, I know that I should've done it that way
 * in the first place ;)).
 *
 * Split up some modules (misc.c, copy_pkts.c). filter_html.c
 * is much more cleaner now. I think that everyone should be able
 * to understand how the HTML parser works now. Especially
 * the skipping of comments is easy to get now ;)
 *
 * This seems to be getting cleaner...
 *
 */

/*
 * a_quote
 *
 * Is there a [HTML-]quote? If so, return the ASCII-quote.
 *
*/

#include "parse_tools.h"
#include "log.h"

/* random seed for threads */
extern unsigned int rand_r_seed; /* from httpf.c */

#ifdef linux
extern sem_t rand_sema;
#endif

/* This will be inserted when we received something declared as text/html containing binary data */
char *binary_warning="<html><head><title>httpf message</title></head><body  bgcolor=\"#ffffff\">"
"\n<p><strong><center>Server sent data with <acronym>MIME</acronym>-type <code>text/html</code> "
"but document contains binary data.</center></strong></p>"
"<p>Since <acronym>html</acronym> must not contain binary data this is considered an error."
"The HTTP-server or its maintainer did something wrong. The document cannot be displayed for security reasons.</p>"
"<p>This is a message from the security proxy <a href=\"http://httpf.sourceforge.net\"><code>httpf</code></a> that monitors the flow of data sent to and from your browser.</p>"
"The following bytes are a sample of the data that failed to pass the proxy."
"The suspicious character is printed in bold face. At the left side you see a hex dump,"
" on the right side the contents are printed in human readable form (if possible). Two dots"
" indicate that the character at that position could not be printed."
"<p>";


char a_quote(char *string) {
    if (strncmp(string, "\"", 1)==0)
	return string[0];
        
    if (strncmp(string, "\'", 1)==0)
	return string[0];
        
    if (strncmp(string, "&quot;", 6)==0)
      return '\"';
    
    if (strncmp(string, "&#39;", 5)==0)
      return '\'';
    
    if (strncmp(string, "&#34;", 5)==0)
	return '\"';
    
    return 0;
}

/*
 * skip_quote
 *
 * Take address where a quote starts and return the address
 * where it ends. We don't check if there's quote.
 */
char *skip_quote (char *s) {
  if (s[0]=='&') {
	if (s[1]=='q')
	  /* &quot; */
	  return s+6;
	else
	  /* &#34; and &#39; */
	  return s+5;
  } else {
	/* single and double quote */
	return s+1;
  }
}


/*
 * find_end
 *
 * Find the end of a correctly nested string.
 * TRAILER is the char the string has to end with.
 * Gets address of the last character (the trailer).
 * Returns:
 *  NULL - invalid attribute. Quotes not nested properly so
 *    we reached the end of the input buffer.
 *  other - addr. of last char.
 * Note: This would be the place to insert a check for the
 *   maximum length of strings.
 */
char *find_end(char *start, char trailer, TCT *tct) {
  char *pos;
  char c;
    
  pos = start;

  logmsg(LOG_DEBUG, "[%d] Looking for trailing %c in attribute value",
		 tct->tid, trailer); 
    
  /* as long as we are in the string */
  while((pos[0]!=0) && (pos < tct->inbuf+tct->inlen)) {
	c = a_quote(pos); /* quote? */
	if (c!=0) {
	  /* we have a quote */
	  switch(trailer) {
	  case '\"':
		if (c=='\'') {
		  pos+=1; /* skip it */
		} else {
		  if (c=='\"') {
			/* match! */
			return skip_quote(pos);
		  }
		}
		break;
		
	  case '\'':
		if (c=='\"') {
		  pos+=1; /* skip it */
		} else {
		    if (c=='\'') {
			  /* match! */
			  return skip_quote(pos);
		    }
		}
		break;
	  }
	} else {
	  /* we don't have a quote */
	  
	  /* gg 18.08.99: proceed in the buffer only if we didn't encounter
		 a quote. Otherwise things like 
		 ... window.status=''">
		 won't be parsed correctly since we happily skip the closing double quote.
		 Bah. This was a naughty bug. :( */
	  pos++;
	}
  }
  
  /* if we are here we didn't find a matching quote. */
  logmsg(LOG_DEBUG, "[%d] End of buffer whilst looking for closing quote", tct->tid);
  tct->error=NEED_MORE_INPUT;
  return NULL;
}


/*
 * find_unquoted_end
 *
 * Find end of unquoted value. This should be either a blank
 * or the end of the string.
 * Returns:
 *   address behind the last character of the value.
 *   0 if the value is illegal
 *
 * Note: the HTML 4.0 spec says an unquoted value may consist of
 * decimals, letters, hyphens and periods.
 * (3.2.2 Attributes HTML 4.0 Specification)
 */
char *find_unquoted_end(char *val_start,TCT *tct) {
  int i=0;

  /* skip legal characters */
  while(legal_val_constituent(val_start[i])) {
    i+=1;
  }

  /* whitespace and '>' are okay */
  if (legal_whitespace(val_start[i]) || val_start[i]=='>')
    return val_start+i;
  else
    /* illegal character encountered, error */
    tct->error=UNALLOWED_CHAR;
    return NULL;
}

/*
 * legal_constituent
 *
 * Takes character and tests if this is a legal constituent
 * for a tag or unquoted attribute, i.e. if it's a letter, digit,
 * period or hyphen.
 * Due to the real world (e.g. yahoo) we allow the slash '/' and the
 * colon ':', too. And '#' (setting colours), '%' (widths etc.) and '?'
 * (cgi), ',' (in frameset: "ROWS=20%,50%,30%" :|).
 * Returns character if legal, 0 otherwise.
 * Note: since we allow '&' we MUST NOT allow '#'... if we did we would
 * allow character references...
 * Do not allow parens... these are usually used in code :)
 */
int legal_constituent(int c) {
  if (isalnum(c) || c=='.' || c=='-' || c=='/' || c==':' || c=='&' || c=='%' || c=='_' || c=='?' || c==',' || c=='@' || c=='~' || c=='+')
	return c;
  else
	return 0;
}

int legal_val_constituent(int c) {
  if (legal_constituent(c) || c=='=')
	return c;
  else
	return 0;
}

/*
 * legal_whitespace
 *
 * Takes character and tests if this is a legal whitespace.
 * Basically it's a isspace() wrapper ;)
 * Returns character if legal, 0 otherwise.
 */
int legal_whitespace(int c)
{
    if (isspace(c))
	return c;
    else
	return 0;
}


/*
 * skip_leading_whitespace
 *
 * Skip legal whitespace. Returns pointer to string or
 * NULL if there was an error. Check tct->error.
 */
char *skip_leading_whitespace(char *buf,TCT*tct) {
  int i=0;
  
  if (check_inbuf_pointer(buf,tct)!=NO_ERROR) return NULL;

  /* skip leading blanks while we are in the string*/
  while(buf[i] && legal_whitespace(buf[i])) {
	i+=1;
  }

  if (check_inbuf_pointer(buf+i,tct)!=NO_ERROR) return NULL;

  return buf+i;
}

/*
 * skip_to_next_whitespace()
 *
 * Proceed to the next whitespace character.
 * Returns pointer to whitespace character or
 * NULL if there's an error. Check tct->error.
 */
char *skip_to_next_whitespace(char *buf,TCT*tct) {
  int i=0;

  if (check_inbuf_pointer(buf,tct)!=NO_ERROR) return NULL;

  /* skip characters while we are in the string*/
  while(buf[i] && !legal_whitespace(buf[i])) {
	i+=1;
  }

  if (check_inbuf_pointer(buf+i,tct)!=NO_ERROR) return NULL;

  if (buf[i]==0) {
    tct->error=UNALLOWED_CHAR;
    return NULL;
  }

  return buf+i;
}


/*
 * check_inbuf_pointer()
 *
 * Check if the given pointer really points to a valid
 * location in the input buffer.
 * thread attribute ERROR will be set accordingly if
 * this is not the case.
 * ERROR will be set to NO_ERROR if pointer is valid.
 */
int check_inbuf_pointer(char*buf,TCT *tct) {

  if (buf==(tct->inbuf+tct->inlen)) {
    logmsg(LOG_INFO,"[%d] At end of buffer", tct->tid);
    tct->error=NEED_MORE_INPUT;
  } else if (buf>tct->inbuf+tct->inlen) {
    logmsg(LOG_ERR,"[%d] BUFFER OVERRUN. THIS IS A BUG. (sorry for yelling ;))",
	   tct->tid);
    tct->error=BUFFER_OVERRUN;
  } else if (buf==NULL) {
    logmsg(LOG_ERR,"[%d] NULL POINTER. THIS IS A BUG. (sorry for yelling ;))",
	   tct->tid);
    tct->error=NULL_POINTER;
  } else if (buf<tct->inbuf) {
    logmsg(LOG_ERR,"[%d] BUFFER UNDERRUN. THIS IS A BUG. (sorry for yelling ;))",
	   tct->tid);
    tct->error=BUFFER_UNDERRUN;
  } else
    tct->error=NO_ERROR;

  return tct->error;
}

/*
 * skip_comment
 *
 * Skips HTML comment. This now a little tidier than before.
 * This flips between the states PARSE_COMMENT and
 * PARSE_COMMENT1 unless we reach the end of comment
 * or the end of the buffer is reached.
 * Notes:
 * We expect to be called with the parse state
 * PARSE_COMMENT or PARSE_COMMENT1. If this is not the case
 * we write a bug message to the log, set
 * parse state to PARSE_CONTENT and return with FAILURE.
 *
 * SUG: imagine this content:
 *   '<!---- ...bla bla... ------ ...bla bla...---->'
 *        x                    x                 x
 *   Skip to the marked positions at (1).
 */
int skip_comment(TCT *tct) {
  int found=0;
  char *hyphen, *ws;
  int rest_buf_len;

  while (!found) {
    switch(tct->parse_state) {
    case PARSE_COMMENT:
      while(tct->parse_state==PARSE_COMMENT) {
	rest_buf_len = tct->inlen-tct->inoffset;
	/* find '--'. This is safe since inbuf is null terminated. */
	hyphen=strstr(tct->inbuf+tct->inoffset, "--");
	if (hyphen != NULL) { /* this is the end of a comment */
	  tct->inoffset=hyphen+2-tct->inbuf; /* point right behind 2nd '-' */
	  logmsg(LOG_DEBUG, "[%d] Found closing -- at offset %d.", tct->tid,hyphen-tct->inbuf);
	  tct->parse_state=PARSE_COMMENT1; /* found -- */
	} else { /* read more data and come back */
	  tct->error=NEED_MORE_INPUT;
	  tct->inoffset=tct->inlen;
	  return FAILURE;
	}
      }
      /* slip through */
    case PARSE_COMMENT1:
      while (tct->parse_state==PARSE_COMMENT1) {
	/* skip possible WS */
	ws = skip_leading_whitespace(tct->inbuf+tct->inoffset,tct);
	if (tct->error!=NO_ERROR) {
	  if (tct->error!=NEED_MORE_INPUT) /* something strange happened, back up */
	    tct->parse_state=PARSE_COMMENT;
	  /* read more data and come back */
	  tct->error=NEED_MORE_INPUT;
	  tct->inoffset=tct->inlen;	
	  return FAILURE;
	}
	
	/* WS skipped */
	tct->inoffset=ws-tct->inbuf;
	if (ws[0]=='>') { /* found closing '>' */
	  tct->parse_state=PARSE_CONTENT;
	  tct->inoffset++;
	  found=1;
	} else { /* (1) comment continues, see suggetions above */
	  tct->parse_state=PARSE_COMMENT1;
	  logmsg(LOG_INFO, "[%d] '--' in comment at %d",
		 tct->tid, tct->inoffset);
	  tct->parse_state=PARSE_COMMENT;
	  /* correct offset: we might have found
	   * something like  '...bla --->'
	   */
	  tct->inoffset--;
	}
      }
      break;
    default:
      logmsg(LOG_ERR, "[%d] BUG: skip_comment called from wrong parse state.",
	     tct->tid, tct->inoffset);
      tct->parse_state=PARSE_CONTENT;
      return FAILURE;
    }
  }
  return OKAY;
}

/*
 * skip_to_end_of_tag()
 *
 * Points inoffset behind the closing '>' of a tag.
 * Please note that this needs correctly nested
 * quotes and angles.
 */
int skip_to_end_of_tag(TCT*tct) {
  int i=tct->inoffset-1;

  /* angles might be in quotes... */
  do {
    i++;
    if (tct->inbuf[i]=='\"') tct->quote_count++;
    if (tct->inbuf[i]=='<') tct->angle_count++;
    if (tct->inbuf[i]=='>') tct->angle_count--;

    /* (1)
     * break if there's a '<' that's not in a string.
     * Again this is a Real World effect from Slashdot:
     * <BR><BR<P>content
     * Yes, otherwise the document will be skipped to the end ;(
     * Note: "tct->angle_count==2" because we're already
     *   counting an opening '<'.
     */
    if ((tct->angle_count==2) && (tct->quote_count%2==0)) {
      logmsg(LOG_NOTICE, "[%d] Real World 2", tct->tid);
      break;
    }

  } while (tct->inbuf[i] && !(tct->angle_count==0 && (tct->quote_count%2==0)));

  if (check_inbuf_pointer(tct->inbuf+i,tct)!=NO_ERROR) {
    tct->inoffset=tct->inlen;
    tct->error=NEED_MORE_INPUT;
    return FAILURE;
  } else {
    if ((tct->inbuf[i])=='>')
      tct->inoffset=i+1; /* skip the closing '>' */
    else /* we're coming from (1) */
      tct->inoffset=i;

    return OKAY;
  }
}



int copy_doctype(char *buf, TCT *tct) {
  logmsg(LOG_DEBUG, "[%d] Copying doctype",
	 tct->tid);
  tct->inoffset++; /* skip the opening '<' */
  /* we're in a tag */
  tct->angle_count=1; tct->quote_count=0;
  /* skip over tag */
  if (skip_to_end_of_tag(tct)==FAILURE)
    return FAILURE; /* leave if doctype exceeds buffer */

  /* copy it */
  if(output_string(buf, (tct->inbuf+tct->inoffset)-buf, tct)==FAILURE) return FAILURE;

  return OKAY;
}

/*
 * inbuf_contains_binary()
 *
 * Check if the input buffer contains binary characters.
 * To make sure that noone can fool us we check the data at
 * random places.
 *
 * We test NUMBER_OF_CHECKS_FOR_BINARY_DATA times if
 * a character in the buffer is printable or not.
 *
 * KinderUnix doesn't seem to have rand_r, so currently
 * we use rand() and protect it's call with a semaphore.
 *
 * Returns -1 if buffer contents seems to be okay or
 * the position of the charater that's suspicious.
 */
#define NUMBER_OF_CHECKS_FOR_BINARY_DATA 16

int inbuf_contains_binary(TCT *tct) {

  int i;
  int result;

  /* Leave immediately if there is no data to check */
  if (tct->inlen<=tct->inoffset) return NO;
  
  logmsg(LOG_DEBUG, "[%d] Trying to validate MIME-type text/html", tct->tid);

#ifdef linux
  sem_wait(&rand_sema); /* wait for semaphore */
#endif

  for (i=0; i<NUMBER_OF_CHECKS_FOR_BINARY_DATA; i++) {
    /*
     * Get a new random number within the buffer.
     * We mod (tct->inlen-tct->inoffset) because we are at the
     * first chunk of data and have the header from 0..inoffset.
     */
#ifdef linux
    /* linux: use thread-unsafe rand */
    int random_position = rand() % (tct->inlen-tct->inoffset);
#else
    /* solaris: use thread-safe rand_r */
    int random_position = rand_r(&rand_r_seed) % (tct->inlen-tct->inoffset);
#endif 
   char c=tct->inbuf[random_position+tct->inoffset];
  
    if (!isprint(c) && !isspace(c)) {
      logmsg(LOG_ERR, "[%d] text/html validation failed at position %d of %d bytes because of character %d",
	     tct->tid, random_position, tct->inlen, c);
      result = random_position+tct->inoffset;
      break;
    } else {
      result = -1;
    }
  }

#ifdef linux
  sem_post(&rand_sema);
#endif

  return result;
}

void print_binary_data_warning(TCT *tct, int offset) {
  int i=0;
  char buf[10];

  output_string(binary_warning,strlen(binary_warning),tct);

  /* prevent buffer underrun whilst printing the info */
  /*  if (offset < 8) return;
   */

  for (i=0; i < 16; i++) {
    if (i==8) output_string("<strong>", 8, tct);
    sprintf(buf, "%2x ",  tct->inbuf[offset-8+i]);
    output_string(buf, 3, tct);
    if (i==8) output_string("</strong>", 9, tct);
  }

  output_string("   ", 3, tct);
  
  for (i=0; i < 16; i++) {
    char c = tct->inbuf[offset-8+i];

    if (i==8) output_string("<strong>", 8, tct);
    if (isprint(c))
      sprintf(buf, "%c ",  tct->inbuf[offset-8+i]);
    else
      sprintf(buf, "..");

    output_string(buf, 2, tct);
    if (i==8) output_string("</strong>", 9, tct);

  }

  output_string("</pre></body></html>", 21, tct);

}
