/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: process_server.c,v 1.4 2001/05/04 12:09:27 gellert Exp $
 *
 * $Log: process_server.c,v $
 * Revision 1.4  2001/05/04 12:09:27  gellert
 * The mime-type check is now done case-insensitive (with strncasecmp).
 *
 * Revision 1.3  2001/03/03 17:25:51  glauschwuffel
 * The check for binary data declared as text/html is now available on
 * Linux, too (protected unsafe rand() with semaphore).
 *
 * Expanded the error page when binary data is found a bit:
 * the contents surrounding the suspicious byte is now displayed.
 *
 * Revision 1.2  2001/03/01 16:02:49  glauschwuffel
 * Major cleanup, we don't have one function anymore but a set a
 * functions. Reads much better now. Adapted for the changes on read.c
 *
 * We set the indicator response_has_http_header if read_server_header()
 * doesn't tell us that we have no header. Useful for things that depend
 * on an existing header (e.g. the MIME-type check for text/html...)
 *
 * Revision 1.1  2000/06/19 22:06:33  glauschwuffel
 * Major overhaul of the HTML parser. Now we have something like a
 * state machine (yes, I know that I should've done it that way
 * in the first place ;)).
 *
 * Split up some modules (misc.c, copy_pkts.c). filter_html.c
 * is much more cleaner now. I think that everyone should be able
 * to understand how the HTML parser works now. Especially
 * the skipping of comments is easy to get now ;)
 *
 * This seems to be getting cleaner...
 *
 */

#include <sys/types.h>
#include <sys/socket.h> /* send() */
#include <errno.h>

#include "ftype_check.h"
#include "filter_html.h"
#include "analyze_response_header.h"
#include "analyze_header.h"
#include "read.h"
#include "mythreads.h"
#include "log.h"
#include "misc.h"
#include "parse_tools.h"
#include "threads.h"

/* from httpf.c */
extern char *httpf_id;

static void initialize();
static int process_header(TCT*);
static int check_for_body(TCT*);
static int process_body(TCT*);
static int buffer_contains_html(TCT*);


/*
 * process_server
 *
 * receive the header of the http-response. parse it
 * and modify it if necessary. then send it.
 */

int process_server(TCT *tct) {

  initialize(tct);
  if (process_header(tct)==FAILURE) return FAILURE;
  /* outbuf is now filled with the header we want to send */

  if (check_for_body(tct)==FAILURE) return OKAY; /* to have no body is okay */
  if (process_body(tct)==FAILURE) return FAILURE;
    
  logmsg(LOG_INFO,"[%d] Connection closed by server", tct->tid); 
  return OKAY;
}

static void initialize(TCT *tct) {
  /* the MIME-type will be copied from the header of the response */
  tct->mime_type[0] = '\0';
    
  /* we have input since we are at the start of processing data */
  tct->more_input_data=YES;

  /* will be YES when the initial thread received the 'hangup'-signal */
  tct->hangup_flag=NO;
  
  /*
   * reset method: servers don't have methods so a value of 0
   * 'indicates' server processing
   * FIXME: this should be a constant :)
   */
  tct->method = 0;

  /* set to YES iff there's a header in the server response */
  tct->response_has_http_header=NO;
  
  /* reset offsets and lengths of input and output buffer */
  reset_buffers(tct);
}

static int process_header(TCT *tct) {
  /* read header */
  if (read_server_header(tct)==FAILURE) {
    /* check if we received the the hangup flag and return immediately if so */
    if (check_for_hangup_flag(tct)==FAILURE) return FAILURE;

    /*
     * If there is no end of header we assume that we don't have a header at all.
     * (HTTP headers are optional).
     * If there is no header we just log this fact and set the according indicator
     * (Yes, the indicator is set to NO within the initialization above, but you never know...)
     * If there is any other error we couldn't receive a decent response and abort.
     */
    if (tct->error==NO_END_OF_HEADER) {
      tct->response_has_http_header=NO;
      logmsg(LOG_ERR, "[%d] Server response contains no header", tct->tid);
    } else {
      logmsg(LOG_ERR, "[%d] FIXME: We should generate an error page here...", tct->tid);
      logmsg(LOG_ERR, "[%d] Receiving data on server socket failed with errno %d.",
	     tct->tid, errno);
      return FAILURE;
    }
  } else {
    /* we know we have a header */
    tct->response_has_http_header=YES;
    analyze_response_header(tct);
    logmsg(LOG_DEBUG, "[%d] Done with header anlaysis", tct->tid);
  }

  return OKAY;  
}

static int check_for_body(TCT *tct) {

  /*
   * If we are at the end of the input buffer there is
   * either no body or the server didn't send it.
   */
  if (tct->inoffset == tct->inlen) {
    /* we are at the end */
    logmsg(LOG_DEBUG, "[%d] No document body", tct->tid);
    if (send(tct->insocket, tct->outbuf, tct->outoffset, 0)<0) {
      logmsg(LOG_ERR, "[%d] Error on sending server header without body", tct->tid);
      tct->error=SOCKET;
      return FAILURE;
    }

    /* is the body following in the next packet? */
    reset_buffers(tct);
    if (read_server_body(tct) == FAILURE) {
      if (check_for_hangup_flag(tct)==FAILURE) return FAILURE;

      /* leave if reading failed */
      logmsg(LOG_ERR, "[%d] There is no body following after the header",
	     tct->tid);
      /* FIXME: this should be an error page: reading failed */
      return FAILURE;
    } else {
      logmsg(LOG_INFO, "[%d] Got new datagram with body", tct->tid);
      return OKAY;
    }
  }
  
  /* we are not at the end of the input buffer, so we assume that we'll have a body */
  return OKAY;
}

static int process_body(TCT *tct) {
  int result;
  int quit=0; /* for the read-body loop */
  int at_beginning_of_body=YES; /* used for checking body contents */

  logmsg(LOG_DEBUG, "[%d] Am now at %d of %d bytes in inbuf.",
	 tct->tid, tct->inoffset, tct->inlen);

  quit=0;
  result=OKAY;

  /* loop until there's an error or we've finished */
  do {
    if (result==FAILURE) {
      logmsg(LOG_ERR, "[%d] FIXME: We should generate an error page here...", tct->tid);
      logmsg(LOG_ERR, "[%d] Receiving data on server socket failed with errno %d.",
	     tct->tid, errno);
      if (tct->more_input_data==NO) {
	break; /* leave on eof */
      } else {
	return FAILURE;
      }
    }
    
    /* reading was successful  */

    if (buffer_contains_html(tct)==NO) {
            /* no html */

      /* check the MIME-type if we are at the first block */
      if (at_beginning_of_body==YES) {
	at_beginning_of_body=NO;
	logmsg(LOG_DEBUG, "[%d] Checking MIME-type at inbuf offset %d.",
	       tct->tid, tct->inoffset);
	    
	if(type_check(tct)!=0) {
	  /*
	   * (1): The data in the input buffer is not of any configured MIME-type.
	   *      Return immediately and don't send the data.
	   */
	  logmsg(LOG_ERR, "[%d] FIXME: Illegal non-HTML data", tct->tid);
	  logmsg(LOG_ERR, "[%d] (this should be an error page)", tct->tid);
	  return OKAY;
	} else {
	  /*
	   * The MIME-type that's given in the header matches the configuration */
	  logmsg(LOG_DEBUG, "[%d] Non-HTML Data: Copying %d of %d bytes",
		 tct->tid, tct->inlen-tct->inoffset, tct->inlen);
	  output_string(tct->inbuf+tct->inoffset , tct->inlen-tct->inoffset, tct);
	  logmsg(LOG_INFO, "[%d] Non-HTML Data: Sending %d bytes",
		 tct->tid, tct->outlen);
	  if (send(tct->insocket, tct->outbuf, tct->outlen, 0)<0) {
	    return OKAY;
	  }
	      
	  /* FIXME: check if sending succeeded */
	  reset_buffers(tct);
	}
      } else {
	/*
	 * We're somewhere in the body. Since we return immediately if the
	 * MIME-type is not correct (see (1)) we reach this point only if
	 * the MIME-type is correct. So we just copy the data we receive.
	 */
		    
	if (tct->inlen > 0) {
	  logmsg(LOG_INFO, "[%d] Non-HTML Data: Sending %d bytes",
		 tct->tid, tct->inlen);
	  
	  if (send(tct->insocket, tct->inbuf, tct->inlen, 0) < 0) {
	    logmsg(LOG_ERR, "[%d] Sending on client socket failed with errno %d.",
		   tct->tid, errno);
	    tct->error=SOCKET;
	    break; /* quit the loop */
	  }

	  /* FIXME: check if sending succeeded */
	  reset_buffers(tct);
	}
	  
      } /* somewhere in the body */
      
    } else {
      /* we have html */
      if (at_beginning_of_body==YES) {
	int binary_offset;

	at_beginning_of_body=NO;

	/* insert our id  so ppl will know when we processed the page. Make this a command switch? */
	output_string(httpf_id,strlen(httpf_id),tct);

	/*
	 * Check if there is binary data inside the buffer.
	 * Leave if this is the case.
	 *
	 * (Run your Apache with a binary file [e.g. image] named
	 * 'thisisagif.html' to see what happens. Yes, right, Apache
	 * will send the image declared as text/html.)
	 */

	binary_offset = inbuf_contains_binary(tct);
	if (binary_offset != -1) {
	  logmsg(LOG_ERR, "[%d] Input data is declared as text/html but contains binary data", tct->tid);
	  print_binary_data_warning(tct, binary_offset);

	  /* send the outbuf so far */
	  if (send(tct->insocket, tct->outbuf, tct->outlen, 0)<0) {
	    logmsg(LOG_INFO, "[%d] Sending the output buffer failed with errno %d",
		   tct->tid, errno);
	    tct->error=CLIENT_SOCKET;
	    return FAILURE;
	  }

	  break;
	}
      }

      /* Filter it. Leave the loop if there's an error */
      if (filter_html(tct) == FAILURE) {
	quit = 1;
      }
    }

    /* abort the function if there's an error on the client's socket. */
    if (tct->error==CLIENT_SOCKET) {
      logmsg(LOG_NOTICE,"[%d] Connection closed by client", tct->tid);
      return FAILURE;
    }

    /* Read next block of data and check if we received the HUP signal */
    result=read_server_body(tct);
    if (check_for_hangup_flag(tct)==FAILURE) return FAILURE;
    
    if (tct->more_input_data==NO) quit=1;
    
  } while (quit==0);

  return OKAY;
}


/*
 * buffer_contains_html()
 *
 * Does the input buffer contain any HTML data?
 * We assume that his is true if
 *  a) the response has no HTTP header
 *  b) the response has a header and the MIME-type is 'text/html'
 *
 * NB: choice 'a)' causes binary data to be parsed as HTML documents.
 *     This is a security feature and not a bug. We don't want
 *     binary data to pass the proxy unchecked...
 */
static int buffer_contains_html(TCT *tct) {
  if (tct->response_has_http_header==NO) return YES; /* a) */

  if (strncasecmp(tct->mime_type, "text/html", min(strlen(tct->mime_type), 9))==0)
    return YES; /* b) */

  return NO;
}
