/*
    Copyright (C) 2000 DFN-FWL (http://www.fwl.dfn.de/fwl)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

/*
 * $Id: threads.c,v 1.1 2001/03/01 16:07:07 glauschwuffel Exp $
 *
 * threads.c -- thread spanning and control functions
 *
 * $Log: threads.c,v $
 * Revision 1.1  2001/03/01 16:07:07  glauschwuffel
 * The thread stuff from httpf.c went here. Also some things for handling
 * signal were added: SIGHUP is blocked, the worker thread 'handler' is
 * here.
 *
 * Cleanup: use of constants OKAY/FAILURE
 *
 *
 */

#include <stdlib.h> /* malloc */
#include <pthread.h>
#include <semaphore.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>
#include <string.h> /* bcopy */
#include <sys/socket.h> /* inet_ntoa */
#include <netinet/in.h>
#include <arpa/inet.h>
#include <signal.h>
#include <errno.h>

#include "threads.h"
#include "mythreads.h"
#include "log.h"
#include "httpf.h"
#include "copy-pkts.h"
#include "open_conn.h"
#include "args.h"

DEADLOCK deadlock;
extern ARGS args;

int  stop_deadlock_checker;

TCT *tcontext; /* all thread 'contexts' */

void allocate_tcts(int number_of_threads) {
  if ( (tcontext = (TCT *) malloc (number_of_threads * sizeof (TCT)) ) == (TCT *)NULL) {
    logmsg (LOG_ERR, "Not enough memory for malloc");
    exit (1);
  } else {
    /*
     * Initialize threads contexts for the very first time.
     * Set socket sizes and mark them as ready.
     */
    int i;
    for (i=0; i<number_of_threads; i++)	{
      (*(tcontext + i )).tid = i;
      (*(tcontext + i )).stat = READY;
      (*(tcontext + i )).insocket = -1;
      (*(tcontext + i )).ServerSockSize = args.server_socket_size;
      (*(tcontext + i )).ClientSockSize = args.client_socket_size;
      (*(tcontext + i )).CopySize = args.copy_size;
      if ( sem_init (&(*(tcontext + i )).wait, 0, 0) ) {
	logmsg(LOG_ERR, "Not enough resources for sem_init(). Aborting");
	exit (EXIT_FAILURE);
      }
    }
  }
}

void give_job_to_thread(int number_of_threads, int newsockfd) {
  int responder;

 ll1:
  responder = find_thread (number_of_threads);
  if (responder >= 0) {
    logmsg(LOG_NOTICE, "Thread #%d is ready to accept connection on socket %d",
	   responder, newsockfd);
    
    /* we found a thread that's ready to work */
    (*(tcontext + responder)).insocket = newsockfd; /* set socket */
    (*(tcontext + responder)).stat = SCHEDULED;     /* mark for find_thread() */

    /* tell the thread to do the work */
    if (sem_post (&(*(tcontext + responder)).wait)) {
      /* error -> thread doesn't work */
      (*(tcontext + responder)).stat = ERROR;
      logmsg(LOG_ERR, "Thread #%d doesn't accept the semaphore", responder);
      goto ll1;                                     /* einen anderen Thread finden */
    } else { /* no error */
      logmsg(LOG_NOTICE, "Giving job to thread #%d.", responder);
    }
  } else if (responder == -1) {
    /* no more unused threads, have to wait */
    usleep (50); /* wait for 50 us */
    goto ll1;
  } else {
    logmsg (LOG_ERR, "FATAL ERROR: all threads are in ERROR state, aborting");
    exit (EXIT_FAILURE);
  }
}

/* Anfrage beantworten: ist die Hauptfunktion aller Threads */
/* Die Threads identifizieren sich durch ihren TCT selbst, sodass 
 * Messwerte in diesem Struct gespeichert werden koennen.
 */
int process_query (TCT *tct) {
  /* child process */
  struct sockaddr_in	dest_addr;
    
  bcopy((char *) &(args.destination_address), (char *) &dest_addr, sizeof(dest_addr));
  if ((tct->outsocket = open_conn(&dest_addr, tct->ServerSockSize, tct)) < 0) {
    logmsg(LOG_ERR, "[%d] Can't connect to destination", tct->tid);
    close(tct->insocket);
    return FAILURE;
  } else {
    logmsg(LOG_NOTICE, "[%d] Successful connection to destination on socket %d",
	   tct->tid, tct->outsocket);
  }
  
  /* process the request */
  if (copy_pkts(tct) == FAILURE) {
    logmsg(LOG_ERR, "[%d] Processing request failed", tct->tid);
    close(tct->insocket);
    close(tct->outsocket);
    return FAILURE;
  } 
  
  logmsg(LOG_NOTICE, "[%d] Request to %s:%d processed successfully",
	 tct->tid,
	 inet_ntoa(dest_addr.sin_addr),
	 dest_addr.sin_port);
  
  if ((close(tct->insocket))< 0)
    logmsg(LOG_ERR, "[%d] Closing client socket failed with errno %d",
	   tct->tid, errno);
  if ((close(tct->outsocket)) <0)
    logmsg(LOG_ERR, "[%d] Closing server socket failed with errno %d",
	   tct->tid, errno);

  return OKAY;
}

/* Threaddy ist die Initialfunktion eines Threads, der die Kontrolle vom
 * Kernel nach dem Initialisieren eines neuen Threads gegeben wird.
 *
 * Es gibt vier Zustaende in denen sich ein Thread auf der Anwendungsebene
 * befinden kann:
 *
 * READY:       der Thread ist bereit einen (neuen) Auftrag zu bearbeiten.
 * SCHEDULED:   der Thread hat eine neuen Auftrag bekommen.
 * WORKING:     der Thread bearbeitet einen Auftrag.
 * ERROR:       der Thread hat einen Fehler erkannt oder wurde von einem
 *              anderen Thread als fehlerhaft erkannt.
 */
void threaddy (TCT *tct) {
  /* og, 11.02.99: drei puffer anlegen, und nicht malloc benutzen. */
  char inbuffer[HTTP_BUFSIZE];
  char outbuffer[HTTP_BUFSIZE];
  char errorbuffer[HTTP_BUFSIZE];
  char mime_type[LEN_MIMETYPE];

  sigset_t blocked_signals;

  /*
   * Block 'Hangup' Signal. The initial thread deals with
   * this signal *only*.
   */

  sigemptyset(&blocked_signals);
  sigaddset(&blocked_signals, SIGHUP);
  sigprocmask(SIG_BLOCK, &blocked_signals, NULL);

  /* init. buffer addresses */
  tct->inbuf = &inbuffer[0];
  tct->outbuf = &outbuffer[0];
  tct->errbuf = &errorbuffer[0];
  tct->mime_type = &mime_type[0];

  /* gg 02.03.99: initialize offsets */
  tct->inoffset=0;
  tct->outoffset=0;
  tct->erroffset=0;

  /* don't skip document parts */
  tct->skip_flag=0;
  
  while (1) {
    /* repeat until father dies */
    if (sem_wait (&(tct->wait))) {
      /* gentle touchdown in thread nirvana */
      logmsg(LOG_ERR, "[%d] sem_wait Error: pthread_exit", tct->tid);
      tct->stat = ERROR;
      pthread_exit ((void *)NULL);
    }

    tct->stat = WORKING;
    process_query(tct);
    tct->stat = READY;
  } /* while */
}



/* Versucht die angegebene Anzahl an Threads zu starten. Gibt die Anzahl der
 * tatsaechlich gestarteten Threads zurueck.
 */
int span_threads (int threads) {
  int i, anz;
  pthread_attr_t tattr;
  pthread_t tid;
  
  pthread_attr_init(&tattr);
  /* wichtig, damit Betriebsmittel von Threads automatisch nach der 
   * Termination freigegeben werden. Also beispielsweise wenn sich ein
   * Thread aufgrund eines Fehlers selber terminiert.
   */
  pthread_attr_setdetachstate(&tattr, PTHREAD_CREATE_DETACHED);
  
  for (i = 0, anz = 0; i<threads; i++) {
    if (pthread_create(&tid, &tattr, (void *(*)())threaddy, (void *)(tcontext + i)) == 0) {
      /* wieder ein neuer Thread erzeugt */
      anz++;
    } else {
      /* diesen Thread konnten wir nicht starten, er sollte also
       * auch keine Auftraege bekommen...
       */
      (*(tcontext + i)).stat = ERROR;
    }
  } /* for */
  
#ifdef solaris
  /* dem Scheduler unsere Vorstellungen zur Anzahl der gleichzeitig
   * laufenden Threads mitteilen., cb. 01.12.97
   */
  {
    int ret = thr_setconcurrency(anz);
    if (ret) {
      /* Fehlerbehandlung: todo*/
    };
  }
#endif
  
  return anz;
}


/*
 * find_thread()
 *
 * Find a thread that's READY to do some work.
 *
 * Results:
 *   >=  0: number of thread
 *    = -1: no free thread available
 *    < -1: all threads terminated => time to panic
 */
#define TCTP (*(tcontext + i))
int find_thread (int number_of_threads) {
  int tid = -1;
  int i, errnr = 0;
    
  for (i=0; i < number_of_threads; i++) {
    TCT *tct=tcontext+i;

    if (tct->stat == READY) {
      /* tja, wer Spass haben moechte, laesst diese Anweisung weg */
      tct->stat = SCHEDULED;
      tid = i;
      break;
    } else if (tct->stat == ERROR) {
      errnr++;
    }      
  }
  
  if ( (tid == -1) && (errnr == number_of_threads) )
    /* all threads have terminated with an error :( */
    return -2;
  else
    return (tid);
}


static int adjust_open_file_limit () {
#define FDUSED 5 /* listen, log, door, 2 im Sinn ;) */
  struct rlimit maxfiles;
  int files_needed;
  
  getrlimit (RLIMIT_NOFILE, &maxfiles);
  /* insocket/outsocket for each thread plus FDUSED */
  files_needed = 2 * args.number_of_threads + FDUSED;
  
  if (files_needed > maxfiles.rlim_max) {
    logmsg(LOG_WARNING, 
	   "Warning: number of threads (%d) exceeds number of available file descriptors (%d)", 
	   args.number_of_threads,  maxfiles.rlim_max );
    maxfiles.rlim_cur = maxfiles.rlim_max;
  } else if (files_needed > maxfiles.rlim_cur) {
    maxfiles.rlim_cur = files_needed;
  }
  
  if (setrlimit (RLIMIT_NOFILE, &maxfiles)) {
    logmsg( LOG_ERR,
	    "Error: unable to adjust the number of filedescriptors");
    return FAILURE;
  }
  
  return( (int) ((maxfiles.rlim_cur - FDUSED) / 2) );
}



/* cb, 29.04.98
 *
 * Uebergabe eines deadlock_t *pointers auf maximale 
 * Zeit, die ein recv() blockieren darf und auf die
 * Wartezeit, die dieser Thread zwischen den Test 
 * schlafen soll.
 *
 * Terminiert, wenn die globale Variable "stop_deadlock_checker" != 0 ist.
 */
void check_deadlock (void *thread_args) {
  DEADLOCK *dl = (DEADLOCK *)thread_args;
  time_t time_now, difftime=0;
  int i;
    
  do {	
    sleep(dl->sleep);    /* Warten bis zum naechsten Check */

    time_now = time(NULL);	
    for (i=0; i< args.number_of_threads; i++){
      if ((tcontext+i)->insocket && (tcontext+i)->recv_time){
	  difftime = time_now - (tcontext+i)->recv_time;
	  if (difftime >= dl->maxwait){
	      close ((tcontext+i)->insocket);
	      logmsg(LOG_ERR, "Closing socket %d of thread %d",
		     (tcontext+i)->insocket, i);		    
	  }
      }
    }		
  } while (!stop_deadlock_checker);
}


void initialize_deadlock_checker(void) {
  deadlock.sleep = args.deadlock_checker_intervall;
  deadlock.maxwait = args.recv_duration; 
}


pthread_t start_deadlock_checker() {
  pthread_t tid;

  stop_deadlock_checker = 0; /* Thread bleibt aktiv solange dies 0 ist */
  pthread_create(&tid, NULL, (void *(*)())check_deadlock, (void *)&deadlock);
  return tid;
}


/*
 * close_descriptors()
 *
 * Close descriptors for WROKING threads.
 * This will be called in the signal handler.
 */
void close_descriptors() {
  int i;

  logmsg(LOG_NOTICE, "Closing descriptors of worker threads");
  for (i=0; i < args.number_of_threads; i++) {
    TCT *tct=tcontext+i;

    if (tct->stat == WORKING) {
      logmsg(LOG_NOTICE, "Closing descriptors %d and %d of thread %d",
	     tct->insocket, tct->outsocket, tct->tid);
      /* tell the thread that we received a signal */
      tct->hangup_flag = YES;

      close(tct->insocket);
      close(tct->outsocket);
    }      
  }
  
}

/*
 * check_for_hangup_flag()
 *
 * This function is to be called when reading input data fails.
 * (May be called in other places, too, but this makes most sense when
 * trying to read from the network since the signal handler closes the sockets.)
 */
int check_for_hangup_flag(TCT *tct) {
  if (tct->hangup_flag==NO) return OKAY;
  return FAILURE;
}

void initialize_worker_threads() {
  int result;
  
  /* Ressource-Limits fuer File-Descriptoren (Socket!) hochsetzen.
   * darf erst nach mkdaemon geschehen, da setsid Ressourcen wieder auf Default setzt.
   */
  if ((result = adjust_open_file_limit()) < args.number_of_threads) { 
    logmsg (LOG_ERR, "Limiting number of threads to %d", result);
    args.number_of_threads = result;
  }

  if (args.number_of_threads < 1) {
    logmsg (LOG_ERR, "Not enough threads, aborting");
    exit(EXIT_FAILURE);
  }
  
  allocate_tcts(args.number_of_threads); /* get memory for the thread contexts */
  
  if (span_threads (args.number_of_threads) < args.number_of_threads ) {
    logmsg(LOG_ERR, "Unable to start requested number of threads");
  }

}

